/*
 * act123.cpp
 * Contains definitions for checking actions 1-3.
 *
 * Copyright 2004-2005 by Dale McCoy.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include<string>
#include<fstream>
#include<cassert>
#include<errno.h>

using namespace std;

#include"inlines.h"
#include"ExpandingArray.h"
#include"sanity_defines.h"
#include"data.h"

#define MAX_TTD_SPRITE 4894

struct act123{
    void init(){
        act1.init();
        defined2IDs.resize(0);
        act3spritenum=0;
    }
    class Act1{
    public:
        Act1(){init();}
        void init(){spritenum=0;used.resize(0);}
        unsigned int feature,numsets,spritenum;
        bool is_used(int set)const{return used[set];}
        void use(int set){used[set]=true;}
    private:
        Expanding0Array<bool>used;
    }act1;
    class IDarray{
    public:
        IDarray(){init();}
        void init(){used.resize(0);sprite.resize(0);v1C.resize(0);}
        bool is_defined(int id)const{return sprite[id]!=0;}
        bool is_used(int id)const{return used[id];}
        unsigned short defined_at(int id)const{return sprite[id];}
        void define(unsigned int id,unsigned int spritenum,bool checks1C){
            used[id]=false;
            sprite[id]=(unsigned short)spritenum;
            v1C[id]=checks1C;
        }
        bool checks1C(int id)const{return v1C[id];}
        void use(int id){used[id]=true;}
    protected:
        Expanding0Array<bool>used;
        Expanding0Array<bool>v1C;
        Expanding0Array<unsigned short>sprite;
    };
    ExpandingArray<IDarray>defined2IDs;
    uint act3feature,act3spritenum;
    static act123&Instance(){
        static act123 obj;
        return obj;
    }
    static const act123&cInstance(){return act123::Instance();}
private:
    act123(){init();}
    act123(const act123&);
    void operator=(const act123&);
};

int parity(uint x){
    int ret=0;
    while(x){
        ret+=x%2;
        x/=2;
    }
    return ret;
}
void sanity_use_id(int feature,int id){
    act123::Instance().defined2IDs[feature].use(id);
}

void sanity_use_set(int set){
    act123::Instance().act1.use(set);
}

void invalidate_act3(){
    act123::Instance().act3spritenum=0;
}

void CheckSpriteNum(uint num,uint offset,act123::Act1&act1,uint feature,bool&mismatch,bool&hasGround){
    uint sprite=num&0x3FFF;
    if(num&1<<31){
        if(feature!=act1.feature&&!mismatch){
            IssueMessage(ERROR,FEATURE_MISMATCH,1,act1.spritenum);
            mismatch=true;
        }
        if(sprite>=act1.numsets)
            IssueMessage(ERROR,UNDEFINED_SPRITE_SET,offset,sprite,act1.spritenum);
        else act1.use(sprite);
    }else if(sprite>MAX_TTD_SPRITE)
        IssueMessage(ERROR,SPRITENUM_TOO_HIGH,offset);
    hasGround|=((num&1<<30)!=0);
    int translate=num>>14&3,color=num>>16&0x3FFF;
    if(translate==3)IssueMessage(ERROR,INVALID_COLOR_TRANS,offset);
    else if((translate==1||translate==2&&color)&&(color<775||color>790)&&(color<795||color>803))
        IssueMessage(WARNING1,INVALID_COLOR_SPRITE,offset+2,color);
}

bool check_id(uint offset,unsigned int id,act123::IDarray&IDs){
    if(id&0x8000)return false;//callback
    else if(id>>8){
        IssueMessage(ERROR,NEITHER_ID_CALLBACK,id,CID);
        return true;
    }else if(!IDs.is_defined(id)){
        IssueMessage(ERROR,UNDEFINED_ID,offset,id);
        return true;
    }else{
        IDs.use(id);
        return IDs.checks1C(id);
    }
}

void Init123(){act123::Instance().init();}

class VarData{
public:
    VarData():width(0){}
    VarData(int x):width(x){}
    int Load(FILE*);
    uint width,maxparam;
};
struct FeatData{
    FeatData():var80(VarData(1)){}
    ExpandingArray<VarData>globvars,var40,var60;
    ExpandingFillArray<VarData>var80;
    uint last80,featfor82;
};

class Check2v:public GlobalPtr<FeatData>{
    ExpandingArray<VarData>globvars;
public:
    void Check(uint,uint,uint,uint,uint)const;
    INHERIT_ASSIGN(FeatData);
    static Check2v&Instance(){static Check2v obj;return obj;}
private:
    void Load();
    Check2v(){Load();}
    void operator=(const Check2v&);
    Check2v(const Check2v&);
};

int VarData::Load(FILE*pFile){
    uint var;
    var=GetCheckByte(dat2v);
    if((width=GetCheckByte(dat2v))==0xF0&&var==0xFF)
        return-1;
    if(!(width&0x7F))return var;
    assert(width&0x80);
    /*  //Read lots of currently unused bytes.
        CHECK_EOF(min);
        if(min==0xFF){
            uint type;
            bool done=false;
            while(!done){
                CHECK_EOF(type);
                switch(type){
                case'-':
                    fgetc(pFile);
                case '+':
                    fgetc(pFile);
                    break;
                case 'x':
                    done=true;
                    break;
                DEFAULT(Checkv2::VarData::Init,type);
                }
            }
        }else{
            for(;i>0;i--)
                min|=fgetc(pFile)<<(i*8);//ignore EOF--I'll catch that later
            CHECK_EOF(max);
            for(i=temp.width;i>0;i--)
                max|=fgetc(pFile)<<(i*8);//ignore EOF--I'll catch that later
        }
    }
    */
    width&=0x7F;
    if((var&0x60)==0x60)maxparam=GetCheckByte(dat2v);
    return var;
}

void Check2v::Load(){
    FILE*pFile=myfopen(dat2v);
    VarData tempvar;
    int var;
    while(true){
        var=tempvar.Load(pFile);
        if(var==-1)break;
        globvars[var]=tempvar;
    }
    _p=new FeatData[MaxFeature()+1];
    for(uint i=0;i<=MaxFeature();i++){
        _p[i].last80=0xFF;
        _p[i].featfor82=GetCheckByte(dat2v);
        while(true){
            var=tempvar.Load(pFile);
            if(var==-1)break;
            if(tempvar.width==0x80){
                _p[i].last80=var-1;
                break;
            }
            if(var==0xFF&&tempvar.width==0xF0)break;
            if(var&0x80)_p[i].var80[var&0x7F]=tempvar;
            else if((var&0x60)==0x60)_p[i].var60[var&0x3F]=tempvar;
            else if(var&0x40)_p[i].var40[var&0x3F]=tempvar;
            else _p[i].globvars[var]=tempvar;
        }
    }
    fclose(pFile);
}

void Check2v::Check(uint feature,uint type,uint var,uint param,uint shift)const{
    VERIFY(feature<=MaxFeature(),Check2v::DoCheck,feature);
    if((type&3)==2)feature=_p[feature].featfor82;
    if(var&0x80){
        if((var&0x7F)>_p[feature].last80)IssueMessage(ERROR,NONEXISTANT_VARIABLE,var);
        else if(!_p[feature].var80[var&0x7F].width)IssueMessage(WARNING1,NONEXISTANT_VARIABLE,var);
        else if(shift>=_p[feature].var80[var&0x7F].width<<3)IssueMessage(WARNING4,SHIFT_TOO_FAR,var);
    }else if((var&0x60)==0x60){
        if(!_p[feature].var60[var&0x3F].width)IssueMessage(WARNING1,NONEXISTANT_VARIABLE,var);
        else{
            if(shift>=_p[feature].var60[var&0x3F].width<<3)IssueMessage(WARNING1,SHIFT_TOO_FAR,var);
            if(param>=_p[feature].var60[var&0x3F].maxparam)IssueMessage(WARNING1,PARAM_TOO_LARGE,param,var);
        }
    }else if(var&0x40){
        if(!_p[feature].var40[var&0x3F].width)IssueMessage(WARNING1,NONEXISTANT_VARIABLE,var);
        else if(shift>=_p[feature].var40[var&0x3F].width<<3)IssueMessage(WARNING1,SHIFT_TOO_FAR,var);
    }else if(globvars[var].width){
        if(shift>=globvars[var].width*8)IssueMessage(WARNING1,SHIFT_TOO_FAR,var);
    }else{
        if(!_p[feature].globvars[var].width)IssueMessage(WARNING1,NONEXISTANT_VARIABLE,var);
        else if(shift>=_p[feature].globvars[var].width<<3)IssueMessage(WARNING1,SHIFT_TOO_FAR,var);
    }
}

int Check1(const string&data,int spritenum){
    act123::Act1&act1=act123::Instance().act1;
    uint length=(uint)data.length()/2;
    if(act1.spritenum)
        for(unsigned int i=0;i<act1.numsets;i++)
            if(!act1.is_used(i))IssueMessage(WARNING1,UNUSED_SET,i,act1.spritenum);
    act1.init();
    if(CheckLength(length,(extract_byte(data,3)==0xFF)?6:4,INVALID_LENGTH,ACTION,1,ONE_OF,4,6))return 0;
    if(!IsValidFeature(ACT1,act1.feature=extract_byte(data,1)))IssueMessage(ERROR,INVALID_FEATURE);
    act1.numsets=extract_byte(data,2);
    if(!act1.numsets)IssueMessage(WARNING1,NO_SETS,1);
    int numsprites=extract_extended(data,3);
    if(!numsprites&&!IsValidFeature(EMPTY1,act1.feature))IssueMessage(WARNING1,NO_SPRITES,1);
    if((numsprites>4&&(act1.feature==7||act1.feature==9))||(numsprites>8&&act1.feature<4)||
        (numsprites>1&&act1.feature==0x0B))IssueMessage(WARNING1,SET_TOO_LARGE);
    else if(numsprites>4&&numsprites<8&&act1.feature<4)IssueMessage(WARNING1,STRANGE_SET_SIZE);
    act1.spritenum=spritenum;
    return numsprites*act1.numsets;
}

void Check2(const string&data,int spritenum){
    act123::Act1&act1=act123::Instance().act1;
    uint feature=extract_byte(data,1),id=extract_byte(data,2),nument1=extract_byte(data,3),nument2=extract_byte(data,4),
        length=(uint)data.length()/2,i,j;
    if(!IsValidFeature(ACT2,feature)){
        IssueMessage(FATAL,INVALID_FEATURE);
        return;
    }
    bool checks1C=false;
    switch(nument1){
    case 0x80:
    case 0x83:
    {
        bool isRand=false;
        unsigned int prevID=(unsigned)-1,rID;
        nument2=extract_byte(data,6);
        if(parity(nument2)!=1)
            IssueMessage(ERROR,RAND_2_NUMSETS);
        if(nument2==1)
            IssueMessage(WARNING3,ONLY_ONE_CHOICE);
        if(CheckLength(length,7+2*nument2,BAD_LENGTH,"nrand"))break;
        for(i=0;i<nument2;i++){
            rID=extract_word(data,7+2*i);
            check_id(7*i+2,rID,act123::Instance().defined2IDs[feature]);
            if(prevID!=(unsigned)-1)
                isRand|=(prevID!=rID);
            prevID=rID;
        }
        if(!isRand&&nument2!=1)
            IssueMessage(WARNING3,NOT_RANDOM);
        break;
    }
    case 0x81:
    case 0x82:
    case 0x85:
    case 0x86:
    case 0x89:
    case 0x8A:
    {
        uint extract=1<<((nument1>>2)&3),off=3,var,param=0,shift,op;
        bool isvar=false;
        while(true){
            if(((var=extract_byte(data,++off))&0xE0)==0x60)param=extract_byte(data,++off);
            Check2v::Instance().Check(feature,nument1,var,param,(shift=extract_byte(data,++off))&0x1F);
            off+=(shift&0xC0?3:1)*extract;
            isvar|=(var!=0x1A&&var!=0x1C);
            checks1C|=(var==0x1C);
            if(!(shift&0x20))break;
            if((op=extract_byte(data,++off))>0x0D)IssueMessage(ERROR,INVALID_OP,off,op);
        }
        nument2=extract_byte(data,++off);
        if(!isvar)IssueMessage(WARNING4,NOT_VARIATIONAL);
        uint width=2+extract*2,end=++off+width*nument2;
        if(CheckLength(length,end+2,BAD_LENGTH,"nvar"))break;
        unsigned int def=extract_word(data,end),vID;
        for(i=off;i<end;i+=width){
            vID=extract_word(data,i);
            isvar=check_id(i,vID,act123::Instance().defined2IDs[feature]);
            if(extract_variable(data,i+2,extract)>extract_variable(data,i+2+extract,extract)){
                IssueMessage((i==off)?WARNING4:WARNING1,UNREACHABLE_VAR,(i-off)/width);
            }else if(def==vID&&(nument2!=1||!isvar))
                IssueMessage(WARNING3,REUSED_DEFAULT);
        }
        check_id(end,def,act123::Instance().defined2IDs[feature]);
        break;
    }
    default:
        if(nument1&0x80)IssueMessage(FATAL,INVALID_TYPE);
        else if(feature<6||feature==0x0B){
            bool mismatch=false,no1=false;
            if(CheckLength(length,2*(nument1+nument2)+5,BAD_LENGTH,VARS,"nument1","nument2"))break;
            switch(feature){
            case 0x0B:
            case 5:
                if(nument1!=1)IssueMessage(ERROR,INVALID_LITERAL,3,nument1,1);
                if(nument2)IssueMessage(ERROR,INVALID_LITERAL,4,nument2,0);
                break;
            default:
                if(!nument1&&feature!=4)IssueMessage(ERROR,NO_REQD_SETS,LOADED);
                if(!nument2)IssueMessage(ERROR,NO_REQD_SETS,(feature==4)?LOTS:LOADING);
            }
            for(i=0;i<nument1+nument2;i++){
                j=extract_word(data,2*i+5);
                if(j&0x8000);//Callback
                else if(j>>8)IssueMessage(ERROR,NEITHER_ID_CALLBACK,j,SET);
                else if(act1.spritenum==0&&!no1){
                    IssueMessage(ERROR,NO_ACT1);
                    no1=true;
                }else if(feature!=act1.feature&&!mismatch&&!no1){
                    mismatch=true;
                    IssueMessage(ERROR,FEATURE_MISMATCH,1,act1.spritenum);
                }else if(j>=act1.numsets)
                    IssueMessage(ERROR,UNDEFINED_SPRITE_SET,2*i+5,j,act1.spritenum);
                else act1.use(j);
            }
        }else if(feature==7||feature==9){
            uint ground=extract_dword(data,4);
            bool mismatch=false,hasGround=(ground!=0);
            if(ground)CheckSpriteNum(ground,4,act1,feature,mismatch,hasGround);
            if(nument1){
                uint off=7;
                if(CheckLength(length,8+10*nument1,BAD_LENGTH,"num-sprites"))break;
                for(i=0;i<nument1;i++){
                    uint building=extract_dword(data,++off),xoff=extract_byte(data,off+=4),
                        yoff=extract_byte(data,++off),zoff=extract_byte(data,++off);
                    CheckSpriteNum(building,8+10*i,act1,feature,mismatch,hasGround);
                    if(!building)IssueMessage(ERROR,NO_BUILDING_SPRITE,off-6);
                    if(zoff!=0x80){
                        uint x=extract_byte(data,++off),y=extract_byte(data,++off);
                        ++off;//zextent
                        if(xoff>15)IssueMessage(ERROR,TOO_LARGE,"xoff");
                        else if(xoff+x>16)IssueMessage(WARNING2,OFF_TILE,off-5,"xoff+xextent");
                        if(yoff>15)IssueMessage(ERROR,TOO_LARGE,"yoff");
                        else if(yoff+y>16)IssueMessage(WARNING2,OFF_TILE,off-4,"yoff+yextent");
                    }else if(i==0)IssueMessage(ERROR,FIRST_SPRITE_CANNOT_SHARE);
                }
                if(++off!=length)IssueMessage(WARNING2,EXTRA_DATA,off);
                if(!hasGround)IssueMessage(WARNING2,NO_GROUNDSPRITE,NONTRANS);
            }else{
                if(!ground)IssueMessage(ERROR,NO_GROUNDSPRITE,GROUND);
                if(CheckLength(length,17,INVALID_LENGTH,TYPE,BASICSTD2,HOUSE_INSTYTILE,17))break;
                uint building=extract_dword(data,8),xoff=extract_byte(data,12),yoff=extract_byte(data,13),
                    x=extract_byte(data,14),y=extract_byte(data,15);
                if(building){
                    CheckSpriteNum(building,8,act1,feature,mismatch,hasGround);
                    if(xoff>15)IssueMessage(ERROR,TOO_LARGE,"xoff");
                    else if(xoff+x>16)IssueMessage(WARNING2,OFF_TILE,12,"xoff+xextent");
                    if(yoff>15)IssueMessage(ERROR,TOO_LARGE,"yoff");
                    else if(yoff+y>16)IssueMessage(WARNING2,OFF_TILE,13,"yoff+yextent");
                }
            }
        }else if(feature==10){
            if(length<15){
                IssueMessage(FATAL,INVALID_LENGTH,TYPE,PROD2,INDUSTRIES,15);
                break;
            }else if(length>15)
                IssueMessage(WARNING2,INVALID_LENGTH,TYPE,PROD2,INDUSTRIES,15);
            if(extract_byte(data,3)!=0)
                IssueMessage(ERROR,INVALID_LITERAL,3,extract_byte(data,3),0);
            if(extract_byte(data,14)>1)
                IssueMessage(ERROR,INVALID_AGAIN);
        }else
            INTERNAL_ERROR(LOGIC,Check2,feature,feature);
    }
    if(act123::cInstance().defined2IDs[feature].is_defined(id)&&!act123::cInstance().defined2IDs[feature].is_used(id))
        IssueMessage(WARNING1,UNUSED_ID,id,act123::cInstance().defined2IDs[feature].defined_at(id));
    act123::Instance().defined2IDs[feature].define(id,spritenum,checks1C);
}

void Check3(const string&data,int spritenum){
    unsigned int feature=extract_byte(data,1),length=(uint)data.length()/2;
    if(!IsValidFeature(ACT3,feature)){
        IssueMessage(FATAL,INVALID_FEATURE);
        return;
    }
    bool isOverride=((extract_byte(data,2)&0x80)!=0);
    if(isOverride&&!IsValidFeature(OVERRIDE3,feature))IssueMessage(ERROR,INVALID_FEATURE);
    unsigned int numIDs=extract_byte(data,2)&0x7F,numCIDs=extract_byte(data,3+numIDs);
    if(numCIDs&&feature>4)IssueMessage(WARNING1,NO_CARGOTYPES);
    if(!isOverride){
        act123::Instance().act3feature=feature;
        act123::Instance().act3spritenum=spritenum;
    }else if(!act123::Instance().act3spritenum)IssueMessage(ERROR,NO_STD_3);
    if(CheckLength(length,6+numIDs+3*numCIDs,BAD_LENGTH,VARS,"n-id","num-cid"))return;
    unsigned int id,def=extract_word(data,4+numIDs+3*numCIDs),i,j;
    for(i=3;i<3+numIDs;i++)
        CheckID(feature,extract_byte(data,i));
    for(i=4+numIDs;i<4+numIDs+3*numCIDs;i+=2){
        j=extract_byte(data,i);
        if(j==0xFE&&feature!=4)
            IssueMessage(ERROR,INVALID_CARGO_TYPE,i,j);
        i++;
        check_id(i,id=extract_word(data,i),act123::Instance().defined2IDs[feature]);
        if(def==id)
            IssueMessage(WARNING1,REUSED_DEFAULT);
    }
    check_id(4+numIDs+3*numCIDs,def,act123::Instance().defined2IDs[feature]);
}

void final123(){
    ManualConsoleMessages();
    bool header=false;
    if(act123::Instance().act1.spritenum)
        for(unsigned int i=0;i<act123::Instance().act1.numsets;i++)
            if(!act123::cInstance().act1.is_used(i)){
                IssueMessage(WARNING1,UNUSED_SET,i,act123::Instance().act1.spritenum);
                if(!header){
                    IssueMessage(WARNING1,UNEXP_EOF_STD2);
                    header=true;
                }
            }
    int j;
    for(uint i=0;i<=MaxFeature();i++){
        header=false;
        for(j=0;j<256;j++)
            if(act123::cInstance().defined2IDs[i].is_defined(j)&&!act123::cInstance().defined2IDs[i].is_used(j)){
                if(!header){
                    IssueMessage(WARNING1,UNUSED2IDLEAD,i);
                    IssueMessage(WARNING1,UNEXP_EOF_CARGOID,i);
                    header=true;
                }
                IssueMessage(WARNING1,UNUSEDIDFINAL,j,act123::cInstance().defined2IDs[i].defined_at(j));
            }
    }
}
